/**
 * 
 * AM29F040 flash reader/writer. Based on meeprommer.
 * 
 * kernelcrash.com
 *
 * YOU MUST CHANGE THE BOARD CONFIG TO DO A 256 BYTE SERIAL RECEIVE BUFFER!!!!!
 * http://www.hobbytronics.co.uk/arduino-serial-buffer-size
 * 
 * You can use this sketch to read and write to an Am29F040 flash ROM using an
 * Arduino Uno and either two 74HC595s OR two 74LS374s
 * 
 * History
 * 
 *  - based on meeprommer from https://github.com/mkeller0815/MEEPROMMER
 *  - 20160205 - v0.01 - I did not have any 74HC595s so modified the meeprommer code to use
 *                       two 74LS374s to latch the address. I also added in xmodem transfer
 *                       code so that you can transfer a file and write it on the fly. The 
 *                       wiring of CE, OE, and WE to the Arduino is very different to 
 *                       meeprommer
 *                       
 *  - 20160312 - v0.02 - I have some 74HC595s now so have made the code more like the original 
 *                       meeprommer. CE, OE and WE are now wired to A3, A4 and A5 just like 
 *                       the original meeprommer. So if you wired up two LS374s like v0.01
 *                       it will not work with this version until you change the CE, OE
 *                       and WE wiring. But have left in ifdefs so that you can 
 *                       still use 74LS374s if you like. Look in config.h. Define HC595 if 
 *                       you want to use 74HC595s. If you want to use 74LS374s then define 
 *                       LS374. You only want one of these two things defined. There is also 
 *                       an attempt to speed up the IO. If you define FAST_IO it goes a 
 *                       little faster (I've optimised a few bits of the code, but not all). 
 *                       If you defined OLD_IO then it uses the old IO code. 
 *                       Again you only want one of these two defined in config.h
 *                       
 *                       I've added an 'm' md5sum command. Have used scottmac's sketch here; 
 *                       https://github.com/scottmac/arduino/tree/master/md5 as a basis. 
 *                       Seems to work. Not a lot of visual feedback (an arduino is not 
 *                       the fastest md5 calculator in the universe, but it can do a 256K
 *                       md5sum in under 30secs). Basic idea is to be 
 *                       able to check that what you just wrote is indeed on the flash ROM
 *                       
 * Basic wiring
 * 
 * Two possibilities
 * 
 * - Use two 74LS374's or two 74LS273 latches to do the address bus
 * - Use two 74HC595s to generate the address bus as per the original meeprommer
 * 
 * 74LS374 version
 * ===============
 * 
 * D2 to D9 of the arduino uno are wired to the databus D0 to D7 of the AM29F040
 * D2 to D9 also wired to the 'D' inputs of two 74LS374 latches
 * ie
 * 
 * Arduino | LS374
 *     D2 - p3 1D
 *     D3 - p4 2D
 *     D4 - p7 3D
 *     D5 - p8 4D
 *     D6 - p13 5D
 *     D7 - p14 6D
 *     D8 - p17 7D
 *     D9 - p18 8D
 *     
 * For the corresponding Q outputs,
 * One 74LS374 latch is wired to A0 to A7 of the am29f040. This is the LOW ADDRESS LATCH
 * The other 74LS374 is wired to A8 to A15 of the am29f040. This is the HIGH ADDRESS LATCH
 * 
 * A0 of arduino goes to p11 CP on the LOW ADDRESS LATCH
 * A1 of arduino goes to p11 CP on the HIGH ADDRESS LATCH
 * 
 * A3 of arduino goes to CE* on the am29f040
 * A4 of arduino goes to OE* on the am29f040
 * A5 of arduino goes to WE* on the am29f040
 * 
 * D10 of arduino goes to A16 of 29f040
 * D11 of arduino goes to A17 of 29f040
 * D12 of arduino goes to A18 of 29f040
 * 
 * 
 * 74HC595 version
 * ===============
 * 
 * D2 to D9 of the arduino uno are wired to the databus D0 to D7 of the AM29F040
 * 
 * A0 is wired as per meeprommer to DS pin of the low (A7 to A0) 74HC595 
 *     (I will call this the low address 74HC595)
 * A1 is wired as per meeprommer to the ST_CP pin of both 74HC595s
 * A2 is wired as per meeprommer to the SH_CP pin of both 74HC595s
 * 
 * A3 of arduino goes to CE* on the am29f040
 * A4 of arduino goes to OE* on the am29f040
 * A5 of arduino goes to WE* on the am29f040
 * 
 * D10 of arduino goes to A16 of 29f040
 * D11 of arduino goes to A17 of 29f040
 * D12 of arduino goes to A18 of 29f040
 * 
 * The Q0 to Q7 outputs of the low address 74HC595 go to 
 *     A0 to A7 of the flash ROM
 *     
 * The Q0 to Q7 ouptuts of the high address 74HC595 (ie. the other one) 
 *     go to A8 to A15 of the flash ROM (in the meeprommer schematic Q7 
 *     is not connected due to the smaller size of the EEPROM.
 *     
 * refer to the meeprommer schematic for how everything else is wired up. 
 * 
 * Commands   
 *    a nnnnn   - sets the  address bus to hex nnnnn
 *    w nnnnn   - write data to flash at hex address nnnnn using an xmodem-crc transfer
 *    r nnnnn mmmmm - read from flash at address nnnnn for mmmmm bytes (all in hex)
 *    i - identify flash rom 
 *    e - erase flash rom
 *    t nnnnn 0xxyy - write two bytes (xx and yy) to flash at nnnnn (for testing)
 *    m nnnnn mmmmm - do an md5 hash of the content of the 
 *         flash rom starting at nnnnn and being mmmmm bytes long
 *    
 * INSTRUCTIONS
 *
 * Wire up arduino uno as per notes above. 
 * Make sure you've made the necessary changes to support 256byte serial buffers.
 * YEAH YOU HAVE TO DO THAT.
 * Load the sketch in the arduino IDE. Edit the config.h to indicate if you are 
 *    using 74HC595s or 74LS374s and if you want to try FAST_IO (well, slightly
 *    faster IO).
 * Program the sketch using the arduino IDE. Turn off the arduino IDE and start
 * a serial terminal program with 115200 bps and 8N1. Hopefully hit ENTER and you should
 * get a '%' prompt.
 *
 * You can read your am29f040 with
 * 
 *  % r 00000 00080
 *
 * (NOTE: the 00080 number of bytes bit is optional)
 * 
 * You can get the manufacturer and the device code from the flash ROM with
 * 
 *  % i
 *  
 *  Note, It works with the am29f040. It may or may not work on any other flash roms
 *  
 *  You can erase the whole chip with
 *  
 *  % e
 *  
 * But the main command is to do an xmodem transfer from your PC to write to the flash.
 * Lets say you have a binary file on your PC that you want to write to the start of 
 * your flash. Do this
 * 
 *  % w 00000
 * 
 * Then choose the option in your terminal program to start an xmodem transfer. In minicom
 * this is 'ctrl-a s xmodem'. And then choose the file to send and let it transfer. When its
 * done. Hit ENTER a few times to get your prompt back and perhaps use the 'r' command' to 
 * check that something was written.
 * 
 *  % m 00000 08000
 *  
 *  md5sum from 00000 for 08000 bytes (ie. first 32K of the flash ROM). Can take a while. 256KB
 *  takes about 30 secs.
 *  
 *  
 *
 * Orig meeprommer author details
 *
 * $Author: mario $
 * $Date: 2013/05/05 11:01:54 $
 * $Revision: 1.2 $
 *
 * This software is freeware and can be modified, reused or thrown away without any restrictions.
 *
 * Use this code at your own risk. I'm not responsible for any bad effect or damages caused by this software!!!
 *
 **/

#include "config.h"
#include "md5stuff.h"

//#define VERSIONSTRING "MEEPROMMER $Revision: 1.2 $ $Date: 2013/05/05 11:01:54 $, CMD:R,r,w,W,V"
#define VERSIONSTRING "AM29F040 READWRITE (based on MEEPROMMER), CMD:a,w,r,i,e,m,v, version 0.02"

// command line
#define ECHO_ON

//XMODEM stuff
#define POSITION_SOH  0
#define POSITION_PACKET 1
#define POSITION_PACKETSLEFT 2
#define POSITION_PAYLOAD 3
#define POSITION_CRC 131
#define POSITION_CRC2 132

#define ACK 0x06
#define EOT 0x04
#define NAK 0x15


#define LOWADDRESSLATCH  A0
#define HIGHADDRESSLATCH A1
#define CE      A3
#define OE      A4
#define WE   A5

#define FLASHA16   10
#define FLASHA17   11
#define FLASHA18   12


//4040 stuff
#define COUNTERMASK 0x0fff

#define DS      A0
#define LATCH   A1
#define CLOCK   A2

#define PORTC_DS      0
#define PORTC_LATCH   1
#define PORTC_CLOCK   2

#define PORTC_CE      3
#define PORTC_OE      4
#define PORTC_WE      5


// eeprom stuff
// define the IO lines for the data - bus
#define D0 2
#define D1 3
#define D2 4
#define D3 5
#define D4 6
#define D5 7
#define D6 8
#define D7 9



//a buffer for bytes to burn
#define BUFFERSIZE 133
byte buffer[BUFFERSIZE];
//command buffer for parsing commands
#define COMMANDSIZE 32
char cmdbuf[COMMANDSIZE];

unsigned long startAddress, endAddress;
unsigned long  lineLength, dataLength;
int loop_cmd = 0; // set to >0 for prompt for command mode
int position; // ptr in xmodem packet
int xmodem_prestart;
int available;
int bytes;  // Make sure this is not defined in 'loop'!
byte crchigh,crclow;
int total;
unsigned long flashaddress;

//define COMMANDS
#define NOCOMMAND    0
#define VERSION      1
#define SET_ADDRESS  2
#define CLEAR_RAM    3

#define READ_HEX    10
#define READ_BIN    11
#define READ_ITL    12

#define WRITE_HEX   20

#define WRITE_BIN   21
#define WRITE_ITL   22
#define WRITE_XMODEM 23

#define FLASH_IDENTIFY 30
#define FLASH_ERASE 31

#define MD5SUM 40




/*****************************************************************
 *
 *  CONTROL and DATA functions
 *
 ****************************************************************/

// switch IO lines of databus to INPUT state
#ifdef OLD_IO
void data_bus_input() {
   
  pinMode(D0, INPUT);
  pinMode(D1, INPUT);
  pinMode(D2, INPUT);
  pinMode(D3, INPUT);
  pinMode(D4, INPUT);
  pinMode(D5, INPUT);
  pinMode(D6, INPUT);
  pinMode(D7, INPUT);
}
#endif

#ifdef FAST_IO
void      data_bus_input() {
  DDRD = DDRD & 0x03; // leave bits 0 and 1 as is
  DDRB = DDRB & 0xfc; // leave bits 7 to 2 as is. 
}
#endif

#ifdef OLD_IO
//switch IO lines of databus to OUTPUT state
void data_bus_output() {
  pinMode(D0, OUTPUT);
  pinMode(D1, OUTPUT);
  pinMode(D2, OUTPUT);
  pinMode(D3, OUTPUT);
  pinMode(D4, OUTPUT);
  pinMode(D5, OUTPUT);
  pinMode(D6, OUTPUT);
  pinMode(D7, OUTPUT);
}
#endif

#ifdef FAST_IO
void data_bus_output() {
  DDRD = DDRD | 0xfc; // leave bits 0 and 1 as is
  DDRB = DDRB | 0x03; // leave bits 7 to 2 as is. 
}

#endif

#ifdef OLD_IO
//read a complete byte from the bus
//be sure to set data_bus to input before
byte read_data_bus()
{
  return ((digitalRead(D7) << 7) +
          (digitalRead(D6) << 6) +
          (digitalRead(D5) << 5) +
          (digitalRead(D4) << 4) +
          (digitalRead(D3) << 3) +
          (digitalRead(D2) << 2) +
          (digitalRead(D1) << 1) +
          digitalRead(D0));

}
#endif

#ifdef FAST_IO
byte read_data_bus()
{

  byte b = ((PIND & 0xfc) >> 2) | ((PINB & 0x03) << 6);
  return b;
  
}


#endif


#ifdef OLD_IO
//write a byte to the data bus
//be sure to set data_bus to output before
void write_data_bus(byte data)
{
  //2 bits belong to PORTB and have to be set separtely
  digitalWrite(D6, (data >> 6) & 0x01);
  digitalWrite(D7, (data >> 7) & 0x01);
  //bit 0 to 6 belong to bit 2 to 8 of PORTD
  // meeprommer used to just OR PIND ... which is wrong. You need to mask too
  PORTD = (PIND & 0x03) | ( data << 2 );
  //digitalWrite(D0, (data & 0x01));
  //digitalWrite(D1, (data >> 1) & 0x01);
  //digitalWrite(D2, (data >> 2) & 0x01);
  //digitalWrite(D3, (data >> 3) & 0x01);
  //digitalWrite(D4, (data >> 4) & 0x01);
  //digitalWrite(D5, (data >> 5) & 0x01);
  
  
}
#endif

#ifdef FAST_IO
void write_data_bus(byte data)
{
  PORTD = (PORTD & 0x03) | ( data << 2 );
  PORTB = (PORTB & 0xfc) | (data >> 6);  
}

#endif

#ifdef HC595

void set_address_bus(unsigned long address)
{

  //get bits 15 to 8 of address
  byte hi = ((address >>8) & 0xff);
  
  //get bits 7 to 0 of address
  byte low = address & 0xff;

  //disable latch line
  bitClear(PORTC,PORTC_LATCH);

  //shift out highbyte
  fastShiftOut(hi);
  //shift out lowbyte
  fastShiftOut(low);

  //enable latch and set address
  bitSet(PORTC,PORTC_LATCH);

  // Now the top 3 address bits
  digitalWrite(FLASHA16, (address >>16) & 0x01);
  digitalWrite(FLASHA17, (address >>17) & 0x01);
  digitalWrite(FLASHA18, (address >>18) & 0x01);
  

}

//faster shiftOut function then normal IDE function (about 4 times)
void fastShiftOut(byte data) {
  //clear data pin
  bitClear(PORTC,PORTC_DS);
  //Send each bit of the myDataOut byte MSBFIRST
  for (int i=7; i>=0; i--)  {
    bitClear(PORTC,PORTC_CLOCK);
    //--- Turn data on or off based on value of bit
    if ( bitRead(data,i) == 1) {
      bitSet(PORTC,PORTC_DS);
    }
    else {      
      bitClear(PORTC,PORTC_DS);
    }
    //register shifts bits on upstroke of clock pin  
    bitSet(PORTC,PORTC_CLOCK);
    //zero the data pin after shift to prevent bleed through
    bitClear(PORTC,PORTC_DS);
  }
  //stop shifting
  bitClear(PORTC,PORTC_CLOCK);
}

#endif // HC595

#ifdef LS374
void set_address_bus(unsigned long address)
{

   //Serial.println("set_address_bus");
   //Serial.println(address);
   data_bus_output();
   byte lowbyte = (address & 0xff);

   write_data_bus(lowbyte);
   delayMicroseconds(1);
   // If you wanted to use a 74LS373 and you were using its 'E' input to latch, you would write a HIGH here,
   // then write a LOW a few lines down.
   digitalWrite(LOWADDRESSLATCH,LOW);
   delayMicroseconds(1);
   // latch address bits
   digitalWrite(LOWADDRESSLATCH,HIGH);

   byte highbyte = ((address >>8) & 0xff);
   //Serial.println("high");
   //Serial.println(highbyte);
   data_bus_output();
   write_data_bus(highbyte);
   delayMicroseconds(1);
   // If you wanted to use a 74LS373 and you were using its 'E' input to latch, you would write a HIGH here,
   // then write a LOW a few lines down.
   digitalWrite(HIGHADDRESSLATCH,LOW);
   delayMicroseconds(1);
   // latch address bits
   digitalWrite(HIGHADDRESSLATCH,HIGH);

 
  digitalWrite(FLASHA16, (address >>16) & 0x01);
  digitalWrite(FLASHA17, (address >>17) & 0x01);
  digitalWrite(FLASHA18, (address >>18) & 0x01);
  
}
#endif

//short function to set the OE(output enable line of the eeprom)
// attention, this line is LOW - active
#ifdef OLD_IO
void set_oe (byte state)
{
  digitalWrite(OE, state);
}
#endif

#ifdef FAST_IO
void set_oe (byte state)
{
  if (state==LOW) {
     PORTC &= ~_BV(PORTC_OE);
  } else {
     PORTC |= _BV(PORTC_OE);
  }
}
#endif

#ifdef OLD_IO
//short function to set the CE(chip enable line of the eeprom)
// attention, this line is LOW - active
void set_ce (byte state)
{
  digitalWrite(CE, state);
}
#endif

#ifdef FAST_IO
void set_ce (byte state)
{
   if (state==LOW) {
     PORTC &= ~_BV(PORTC_CE);
  } else {
     PORTC |= _BV(PORTC_CE);
  }
}
#endif

#ifdef OLD_IO
//short function to set the WE(write enable line of the eeprom)
// attention, this line is LOW - active
void set_we (byte state)
{
  digitalWrite(WE, state);
}
#endif

#ifdef FAST_IO
void set_we (byte state)
{
   if (state==LOW) {
     PORTC &= ~_BV(PORTC_WE);
  } else {
     PORTC |= _BV(PORTC_WE);
  }
}
#endif


//highlevel function to read a byte from a given address
byte read_byte(unsigned long address)
{
  byte	    data = 0;

  set_address_bus(address);
  data_bus_input();
  //first disable output
  set_oe(HIGH);
  //disable write
  set_we(HIGH);
  //enable chip select
  set_ce(LOW);
  
  
  //enable output
  set_oe(LOW);
  delayMicroseconds(1);
  data = read_data_bus();

  //disable output
  set_oe(HIGH);
  set_ce(HIGH);
  return data;

}





//highlevel function to read a byte from a current address of 4040 and then bump it up one
byte read_next_byte(void)
{
  byte	    data = 0;

  //Serial.println("read_next_byte");
  set_address_bus(flashaddress);
  
  //set databus for reading
  data_bus_input();
  //first disable output
  set_oe(HIGH);
  //disable write
  set_we(HIGH);
  //enable chip select
  set_ce(LOW);
  //enable output
  set_oe(LOW);
  delayMicroseconds(1);

  data = read_data_bus();

  //disable output
  set_oe(HIGH);
  set_ce(HIGH);

  flashaddress++;

  return data;

}


byte write_next_byte(byte data)
{
   int i;
   byte tmp;
   
   special_flash_write(0x5555,0xaa);
   special_flash_write(0x2aaa,0x55);
   special_flash_write(0x5555,0xa0);
   
   // write the byte
   special_flash_write(flashaddress,data);

   // wait for DQ7 to 
   for (i=0;i<10;i++) {
      tmp = read_byte(flashaddress);
      if (tmp == data)
         break;
   }
   
   flashaddress++;


}

// Send the special address/command combos. The address is written on before WE* falls
// and the command byte b is written before the WE* rises.
void special_flash_write(unsigned long address, byte b)
{
  byte      data = 0;

  //Serial.println("read_next_byte");
  set_address_bus(address);
  
  data_bus_output();
  set_ce(HIGH);
  set_oe(HIGH);
  set_we(HIGH);

  set_ce(LOW);
  set_we(LOW); //latch the address
  delayMicroseconds(1);
  write_data_bus(b);
  set_we(HIGH); // latch data
  set_ce(HIGH);
  data_bus_input();

}


void identify_chip(void) {

   special_flash_write(0x0000, 0xf0); // read-reset
   

   // autoselect
   special_flash_write(0x5555,0xaa);
   special_flash_write(0x2aaa,0x55);
   special_flash_write(0x5555,0x90);
   delayMicroseconds(1);

   byte manf_code = read_byte(0x0000);
   byte device_code = read_byte(0x0001);
   
   Serial.print("manufacturer code: ");
   printByte(manf_code); Serial.println("");
   Serial.print("device code: ");
   printByte(device_code); Serial.println("");

   if ((manf_code == 0x01) && (device_code == 0xa4)) {
       Serial.println("Detected Am29F040");
   }
   special_flash_write(0x0000, 0xf0); // read-reset
   
}

void flash_chip_erase(void) {
   Serial.println("Erase entire chip");
   
   special_flash_write(0x0000, 0xf0); // read-reset

   // autoselect
   special_flash_write(0x5555,0xaa);
   special_flash_write(0x2aaa,0x55);
   special_flash_write(0x5555,0x80);
   special_flash_write(0x5555,0xaa);
   special_flash_write(0x2aaa,0x55);
   special_flash_write(0x5555,0x10);

   int i;
   byte tmp;
   for (i=0;i<50;i++) {
      tmp = read_byte(0x0000);
      if (tmp & 0x80) {
        Serial.println("");
        Serial.println("Erase complete");
        break;
      }
      Serial.print(".");
      delay(1000);
   }
}

/************************************************
 *
 * COMMAND and PARSING functions
 *
 *************************************************/

//waits for a string submitted via serial connection
//returns only if linebreak is sent or the buffer is filled
void readCommand() {
  //first clear command buffer
  for (int i = 0; i < COMMANDSIZE; i++) cmdbuf[i] = 0;
  //initialize variables
  char c = ' ';
  int idx = 0;
  //now read serial data until linebreak or buffer is full
  while (1) {
    if (idx >= COMMANDSIZE) {
      cmdbuf[idx - 1] = 0;
      break;
    }
    if (Serial.available()) {
      c = Serial.read();
      if ((c == '\n') || (c == '\r')) {
        cmdbuf[idx] = 0;
        
        break;
      }
#ifdef ECHO_ON      
      Serial.write(c); // echo it back
#endif
      cmdbuf[idx++] = c;
    }
  }
}

//parse the given command by separating command character and parameters
//at the moment only 5 commands are supported
byte parseCommand() {
  //set ',' to '\0' terminator (command string has a fixed strucure)
  //first string is the command character
  cmdbuf[1]  = 0;
  //second string is startaddress (5 bytes)
  cmdbuf[7]  = 0;
  //third string is a data length (orig file said endaddress (5 bytes))
  cmdbuf[13] = 0;
  //fourth string is length (2 bytes)
  cmdbuf[16] = 0;
  startAddress = hexFiveWord((cmdbuf + 2));
  dataLength = hexFiveWord((cmdbuf + 8));
  lineLength = hexByte(cmdbuf + 14);
  byte retval = 0;
  switch (cmdbuf[0]) {
    case 'a':
      retval = SET_ADDRESS;
      break;
    case 'c':
      retval = CLEAR_RAM;
      break;
    case 'r':
      retval = READ_HEX;
      break;
    case 'w':
      retval = WRITE_XMODEM;
      break;
    case 't':  // write TWO bytes t xxxxx 0aabb
      retval = WRITE_HEX;
      break;  
    case 'v':
      retval = VERSION;
      break;
    case 'i':
      retval = FLASH_IDENTIFY;
      break;
    case 'e':
      retval = FLASH_ERASE;
      break;
    case 'm':
      retval = MD5SUM;
      break;
    default:
      retval = NOCOMMAND;
      break;
  }

  return retval;
}

/************************************************************
 * convert a single hex digit (0-9,a-f) to byte
 * @param char c single character (digit)
 * @return byte represented by the digit
 ************************************************************/
byte hexDigit(char c)
{
  if (c >= '0' && c <= '9') {
    return c - '0';
  }
  else if (c >= 'a' && c <= 'f') {
    return c - 'a' + 10;
  }
  else if (c >= 'A' && c <= 'F') {
    return c - 'A' + 10;
  }
  else {
    return 0;   // getting here is bad: it means the character was invalid
  }
}

/************************************************************
 * convert a hex byte (00 - ff) to byte
 * @param c-string with the hex value of the byte
 * @return byte represented by the digits
 ************************************************************/
byte hexByte(char* a)
{
  return ((hexDigit(a[0]) * 16) + hexDigit(a[1]));
}

/************************************************************
 * convert a hex word (0000 - ffff) to unsigned int
 * @param c-string with the hex value of the word
 * @return unsigned int represented by the digits
 ************************************************************/
unsigned int hexWord(char* data) {
  if (data[0]==0) {
     return 0;
  }
  return ((hexDigit(data[0]) * 4096) +
          (hexDigit(data[1]) * 256) +
          (hexDigit(data[2]) * 16) +
          (hexDigit(data[3])));
}

// 5 digit hex number
unsigned long hexFiveWord(char* data) {
  if (data[0]==0) {
     return 0;
  }
  return ((hexDigit(data[0])*65536L) +
          (hexDigit(data[1]) * 4096L) +
          (hexDigit(data[2]) * 256L) +
          (hexDigit(data[3]) * 16L) +
          (hexDigit(data[4])));
}
/************************************************
 *
 * INPUT / OUTPUT Functions
 *
 *************************************************/


/**
 * read a data block from eeprom and write out a hex dump
 * of the data to serial connection
 * @param from       start address to read fromm
 * @param to         last address to read from
 * @param linelength how many hex values are written in one line
 **/
void read_block(unsigned long  from, unsigned long  to, int linelength)
{
  flashaddress=from;
  set_address_bus(flashaddress);
  //count the number fo values that are already printed out on the
  //current line
  int		    outcount = 0;
  //loop from "from address" to "to address" (included)
  for (unsigned long address = from; address <= to; address++) {
    if (outcount == 0) {
      //print out the address at the beginning of the line
      //Serial.println();
      Serial.print("\r\n");
      Serial.print("0x");
      printAddress(address);
      Serial.print(" : ");
    }
    //print data, separated by a space
    byte data = read_next_byte();
    printByte(data);
    Serial.print(" ");
    outcount = (++outcount % linelength);

  }
  //print a newline after the last data line
  Serial.println();

}

void read_md5(unsigned long startAddress, unsigned long dataLength) {
  unsigned long  bytesleft;
  int bytestoprocess;
  int i;
  char md5buffer[64];
  char md5str[33];
  MD5_CTX context;
  unsigned char digest[16];
  
  flashaddress=startAddress;
  set_address_bus(flashaddress);

  md5str[0] = '\0';
  MD5Init(&context);
  
  while (flashaddress < (startAddress + dataLength) ) {
     // Work out how many bytes are left
     bytesleft = ((startAddress + dataLength) - flashaddress);
     bytestoprocess=64;
     if (bytesleft < 64) {
        bytestoprocess = bytesleft;
     }
     for (i=0;i<bytestoprocess;i++) {
        md5buffer[i] = (char) read_next_byte();
     }
     MD5Update(&context, md5buffer, bytestoprocess);
     // No need to bump up flash address as read_next_byte does it.
  }

  MD5Final(digest, &context);

  make_digest(md5str, digest, 16);
  Serial.println(md5str);
  Serial.println();
  
}
// debug function
void write_two_bytes(unsigned long from, byte a, byte b)
{
  // prep
  special_flash_write(0x0000, 0xf0); // read-reset
  
  flashaddress=from;
  write_next_byte(a);
  write_next_byte(b);
  //print a newline after the last data line
  Serial.println();

}

/**
 * read a data block from eeprom and write out the binary data
 * to the serial connection
 * @param from       start address to read fromm
 * @param to         last address to read from
 **/
void read_binblock(unsigned long from, unsigned long to) {
  for (unsigned long address = from; address <= to; address++) {
    Serial.write(read_byte(address));
  }
  //print out an additional 0-byte as success return code
  Serial.print('\0');

}

/**
 * write a data block to the eeprom
 * @param address  startaddress to write on eeprom
 * @param buffer   data buffer to get the data from
 * @param len      number of bytes to be written
 **/
void write_block(unsigned long address, byte* buffer, int len) {
  flashaddress=address;

  for (unsigned int i = 0; i < len; i++) {
    write_next_byte(buffer[i]);
  }
}


/**
 * print out a 16 bit word as 4 character hex value
 **/
void printAddress(unsigned long address) {
  if (address < 0x0010) Serial.print("0");
  if (address < 0x0100) Serial.print("0");
  if (address < 0x1000) Serial.print("0");
  if (address < 0x10000) Serial.print("0");
  Serial.print(address, HEX);

}

/**
 * print out a byte as 2 character hex value
 **/
void printByte(byte data) {
  if (data < 0x10) Serial.print("0");
  Serial.print(data, HEX);
}



/**
 *  Calc the Xmodem CRC16
 */
unsigned int calccrc(byte *ptr, int count)
{
    unsigned int  crc;
    int i;
    
    crc = 0;
    
    while (--count >= 0)
    {
        crc = crc ^ (unsigned int) (*ptr++ << 8);
        i = 8;
        do
        {
            if (crc & 0x8000)
                crc = crc << 1 ^ 0x1021;
            else
                crc = crc << 1;
        } while(--i);
    }
    return (crc);
}



/************************************************
 *
 * MAIN
 *
 *************************************************/
void setup() {
  data_bus_input();
  flashaddress=0;  // master address for sequential data

  //Turn LED OFF
  //pinMode(LEDPIN, OUTPUT);
  //digitalWrite(LEDPIN, LOW);
#ifdef LS374
  pinMode(LOWADDRESSLATCH,OUTPUT);
  digitalWrite(LOWADDRESSLATCH,HIGH);
  pinMode(HIGHADDRESSLATCH,OUTPUT);
  digitalWrite(HIGHADDRESSLATCH,HIGH);
#endif

#ifdef HC595
  pinMode(DS,OUTPUT);
  pinMode(LATCH,OUTPUT);
  pinMode(CLOCK,OUTPUT);
#endif
  
  pinMode(CE, OUTPUT);
  digitalWrite(CE, HIGH);
  pinMode(WE, OUTPUT);
  digitalWrite(WE, HIGH);

  pinMode(OE, OUTPUT);
  digitalWrite(OE, HIGH);
  
  pinMode(FLASHA16, OUTPUT);
  pinMode(FLASHA17, OUTPUT);
  pinMode(FLASHA18, OUTPUT);


  Serial.begin(115200);
  Serial.println("Started. Note: all hex addresses are five digits");
  Serial.println("  r nnnnn mmmmm - show mmmmm bytes at address nnnnn");
  Serial.println("  w nnnnn - write to flash using xmodem transfer");
  Serial.println("  e - erase flash rom");
  Serial.println("  i - identify flash rom");
  Serial.println("  m nnnnn mmmmm - md5sum rom content starting at nnnnn for mmmmmm bytes long");
  
}







/**
 * main loop, that runs invinite times, parsing a given command and
 * executing the given read or write requestes.
 **/
void loop() {
  byte a, b;

  byte xmodem_byte;
  unsigned int c;

  unsigned int crc;

  if (loop_cmd == NOCOMMAND) {
    Serial.print("\r% "); // show the prompt
    readCommand();
#ifdef ECHO_ON    
    Serial.println(""); // only required in ECHO mode
#endif    
    byte cmd = parseCommand();
    
    switch (cmd) {
      // clear ram is a leftover from when I used some of this code to write to a static ram.
      case CLEAR_RAM:
        Serial.print("Clear ram from 0x");
        Serial.print(cmdbuf + 2);
        Serial.println("");
        set_address_bus(startAddress);
        if (dataLength == 0) {
           dataLength = 2048;
        }
        
        for (c = 0; c < dataLength; c++) {
          write_next_byte(0);
        }
        // reset the address back to the start address
        set_address_bus(startAddress);
        Serial.print("Cleared 0x");
        Serial.print(dataLength,HEX);   
        Serial.println(" bytes");
        break;

      // setting the address like this is mainly used for testing to see if your 74HC595s or 74LS374s are working.
      case SET_ADDRESS:
        // Set the address bus to an arbitrary value.
        // Useful for debugging shift-register wiring, byte-order.
        // e.g. a,00FF
        Serial.print("Setting address bus to 0x");
        Serial.println(cmdbuf + 2);
        Serial.println(startAddress);
        set_address_bus(startAddress);
        
        

        break;
      case FLASH_IDENTIFY:
        Serial.println("Identify FLASH chip");
        identify_chip();
        
        break;
      case FLASH_ERASE:
         flash_chip_erase();
         break;
         
      case READ_HEX:
        //set a default if needed to prevent infinite loop
        if (lineLength == 0) lineLength = 16;
        if (dataLength == 0) {
           dataLength = 0x80;
        }
        endAddress = startAddress + dataLength - 1;
        read_block(startAddress, endAddress, lineLength);
        break;
      case MD5SUM:
        //set a default if needed to prevent infinite loop
        
        if (dataLength == 0) {
           dataLength = 0x80;
        }
        endAddress = startAddress + dataLength - 1;
        Serial.println("Reading md5");
        read_md5(startAddress, dataLength);
        break;
      case READ_BIN:
        endAddress = startAddress + dataLength - 1;
        read_binblock(startAddress, endAddress);
        break;
      case READ_ITL:
        break;
      case WRITE_ITL:
        //take care for max buffer size
        if (dataLength > 1024) dataLength = 1024;
        endAddress = startAddress + dataLength - 1;
        while (bytes < dataLength) {
          if (Serial.available()) buffer[bytes++] = Serial.read();
        }
        write_block(startAddress, buffer, dataLength);
        //Serial.println("Wrote %d bytes",(int) dataLength);

        break;
      case WRITE_HEX: // t xxxxx 0yyzz  Write at xxxxx the two bytes yy and zz
        // simple test write of two bytes at the address
        a = ((dataLength >> 8) & 0xff);
        b = dataLength & 0xff;
        write_two_bytes(startAddress, a, b);
        break;
      case WRITE_XMODEM:
        Serial.println("Please start an xmodem transfer in your terminal ...");
        position = 0;
        bytes = 0;
        total=0;
        flashaddress=startAddress;
        special_flash_write(0x0000, 0xf0); // read-reset
        //set_address_bus(startAddress);
        xmodem_prestart = 1;

        loop_cmd = WRITE_XMODEM;
        break;
      case VERSION:
        Serial.println(VERSIONSTRING);
        break;
      default:
        break;
    }
    
    
  } else if (loop_cmd == WRITE_XMODEM) {
    if (xmodem_prestart) {  // In prestart mode we are just waiting for the sending device to start sending.
      Serial.write('C');
      delay(2000);
      if (Serial.available()) {
        xmodem_prestart = 0;
      }
    } else {
      //should be a byte read to read.
      while (Serial.available()) {
         xmodem_byte = Serial.read();
         switch(position) {
            case POSITION_SOH:
               bytes=0;
               if (xmodem_byte == EOT) {
                  Serial.write(ACK);  // send ACK
                  loop_cmd = NOCOMMAND;
                  //Serial.print("Wrote ");
                  //Serial.println(total,DEC);
               } 
               break;
            case POSITION_CRC:
               crchigh = xmodem_byte;
               break;
            case POSITION_CRC2:
               crc = calccrc(buffer,128);
               if ( (((crc >> 8) &0xff) == crchigh) && ( (crc & 0xff) == xmodem_byte) ) {
                  for (int i=0;i<128;i++) {
                     write_next_byte(buffer[i]);
                     total++;
                  }
                  Serial.write(ACK);  // send ACK
               } else {
                  Serial.write(NAK); // CRC failed. Send NAK, and because we never wrote anything to SRAM, the address counter wont have moved
               }
               position = -1;
               break;
            default:
               if ((position >=POSITION_PAYLOAD) && (position <POSITION_CRC)) {
                  buffer[bytes++] = xmodem_byte;
               }
               break;
          
         }
         position++;
      }  
    }
  }
}





